# -*- coding: utf-8 -*-
# Copyright (c) 2014-2022 Richard Hull and contributors
# See LICENSE.rst for details.

"""
OLED display driver for SSD1306, SSD1309, SSD1322, SSD1325, SSD1327, SSD1331,
SSD1351, SSD1362 and SH1106 devices.
"""

__version__ = '3.10.0'
